#include "cset.hh"

const std::string UTF8seq(unsigned n)
{
    std::string result;
    
    if(n < 0x80)                 // <=7 bits
        result += (char)n;
    else
    {
        if(n < 0x800)            // <=11 bits
            result += (char)(0xC0 + (n>>6));
        else
        {
            if(n < 0x10000)      // <=16 bits
                result += (char)(0xE0 + (n>>12));
            else                 // <=21 bits
            {
                result += (char)(0xF0 + (n>>18));
                result += (char)(0x80 + ((n>>12)&63));
            }
            result += (char)(0x80 + ((n>>6)&63));
        }
        result += (char)(0x80 + (n&63));
    }
    return result;
}

const std::basic_string<unsigned> UTF8toU(const std::string& input)
{
    std::basic_string<unsigned> retval;
    for(unsigned pos=0; pos<input.size(); )
    {
        /* This must be the most obfuscated UTF-8 decoder I have seen.
         * I tried to avoid having many hardcoded magical numbers. -Bisqwit
         */
        unsigned char headbyte = input[pos];
        
        /* How many bytes does each sequence take? */
        static const char sizes[16] =
        { 1,1,1,1,1,1,1,1,  
          0,0,0,0,2,2,3,4 };
        
        /* What is the minimum value of the given sequence? */
        static const unsigned minimums[4] = { 0, 0x80, 0x800, 0x10000 };
        /* How to mask bits from the first byte */
        static const char     masks[4]    = { 0x7F, 0x1F, 0x0F, 0x07 };
        
        unsigned len = sizes[headbyte >> 4];
        
        //fprintf(stderr, "UTF-8 in: %c -> len=%u\n", (char)headbyte, len);
        
        /* Ensure it wasn't an invalid sequence */
        if(len == 0)
        {
        IlSeq:
            retval += '?';
            ++pos; continue;
        }
        /* Ensure we have enough bytes */
        if(pos+len > input.size()) goto IlSeq;
        
        unsigned result=0, shl=0;
        
        /* Process the tail bytes - last to first. */
        for(unsigned n = len; --n > 0; )
        {
            unsigned char byte = input[pos+n];
            
            //fprintf(stderr, "byte %u: %02X\n", n, byte);
            
            // The tail bytes must be 10xxxxxx
            if((byte & 0xC0) != 0x80) goto IlSeq;
            
            unsigned bits = byte & 0x3F;
            result |= bits << shl;      
            shl += 6;
        }
        
        /* Process the head byte. */
        // The top bits have already been verified in sizes[].
        // Take the low bits.
        unsigned bits = headbyte & masks[len-1];
        result |= bits << shl;
        
        if(result < minimums[len-1])
        {
            /* Non-optimal coding - not likely valid. */
            goto IlSeq;
        }

        //fprintf(stderr, "UTF-8 in: %.*s -> %X\n", len, input.c_str()+pos, result);
        
        retval += result;
        pos    += len;
    }
    return retval;
}

bool UTF8_CompareCI(const std::string& s1, const std::string& s2)
{
    /* Test whether the two strings match. */
    /* We apply the easiest method: convert the strings
     * to Unicode strings and then compare them.
     */
    
    std::basic_string<unsigned> s1u = UTF8toU(s1);
    std::basic_string<unsigned> s2u = UTF8toU(s2);

    #define utf8upper(c)   \
       ((c>=0x61 && c<=0x7A) ? (c-0x20) \
      : (c>=0xE0 && c<=0xFE) ? (c-0x20) \
      : (c))
    
    if(s1u.size() != s2u.size()) return false;
    
    for(unsigned a=0; a<s1u.size(); ++a)
        if(utf8upper(s1u[a])
        != utf8upper(s2u[a]))
            return false;

    return true;
}
