#include <algorithm>
#include <iostream>
#include <cctype>

#include "xml.hh"
#include "cset.hh"

#undef isalnum
#undef isspace

namespace
{
    typedef const struct { } xml_error;   // for throwing
    typedef const struct { } xml_end_tag; // for throwing
    
    bool is_tag_char(char c)
    {
        return std::isalnum(c)
            || c == '_' || c == ':'
            || c == '-' || c == '.';
    }
    
    std::string ParseEntity(const std::string& s, unsigned& pos, bool utf8)
    {
        if(s.substr(pos, 5) == "&amp;"    ) { pos += 5; return "&"; }
        if(s.substr(pos, 6) == "&quot;"   ) { pos += 6; return "\""; }
        if(s.substr(pos, 4) == "&lt;"     ) { pos += 4; return "<"; }
        if(s.substr(pos, 4) == "&gt;"     ) { pos += 4; return ">"; }
        if(s.substr(pos, 2) == "&#")
        {
            const char* begin = s.c_str()+pos;
            const char* end;
            
            wchar_t code;
            if(begin[2] == 'x')
            {
                code = std::strtol(begin+3, const_cast<char**> (&end), 16);
            }
            else
                code = std::strtol(begin+2, const_cast<char**> (&end), 10);
            pos += (end-begin);
            if(s[pos] == ';') ++pos;

            std::string result;
            
            if(utf8)
                result += UTF8seq(code);
            else
                result += code;
            
            return result;
        }
        std::string result; result += s[pos++];
        return result;
    }
    
    void ParseSubTree(XMLTree& result, const std::string& s, unsigned& pos,
                      bool KeepNS, bool utf8);
    
    void ParseTag(XMLTree& result, const std::string& s, unsigned& pos,
                  bool KeepNS,
                  bool& utf8)
    {
        std::string tagname;
        bool begin_slash = false;
        bool meta_tag = false;

        //std::cerr << "input (" << s.substr(pos, 100) << ")\n";
        
        if(pos < s.size() && s[pos] == '/') { begin_slash=true; ++pos; }
        if(pos < s.size() && s[pos] == '?') { meta_tag=true; ++pos; }
        if(pos < s.size() && s[pos] == '!')
        {
            ++pos;
            bool is_comment = false;
            for(;;)
            {
                while(pos < s.size() && std::isspace(s[pos])) ++pos;
                if(pos >= s.size()) throw xml_error();
                if(pos+2 <= s.size() && s[pos]=='-' && s[pos+1]=='-')
                {
                    pos += 2;
                    is_comment = !is_comment;
                    continue;
                }
                if(s[pos] == '>' && !is_comment) { ++pos; break; }
                ++pos;
            }
            return;
        }
        
        XMLTreeP subtree = new XMLTree;
        
        while(pos < s.size() && is_tag_char(s[pos]))
        {
            if(!KeepNS && s[pos] == ':') { tagname.clear(); ++pos; continue; }
            tagname += s[pos++];
        }

        //std::wcerr << "tag: " << tagname << "\n";
    
        // Loop until the end of the tag has been reached
        bool found_trailing_slash = begin_slash;
        for(;;)
        {
            while(pos < s.size() && std::isspace(s[pos])) ++pos;
            if(pos >= s.size()) throw xml_error();
            if(meta_tag && s[pos] == '?') { ++pos; continue; }
            if(s[pos] == '>') { ++pos; break; }
            
            if(s[pos] == '/' && !found_trailing_slash)
                { found_trailing_slash=true; ++pos; continue; }
            if(!is_tag_char(s[pos])) throw xml_error();
            
            std::string tagparamname;
            while(pos < s.size() && is_tag_char(s[pos]))
            {
                if(!KeepNS && s[pos] == ':') { tagparamname.clear(); ++pos; continue; }
                tagparamname += s[pos++];
            }
            
            //std::wcerr << "param: " << tagparamname << "\n";
            
            while(pos < s.size() && std::isspace(s[pos])) ++pos;
            if(pos >= s.size()) throw xml_error();
            if(s[pos] != '=') throw xml_error();  ++pos;
            
            while(pos < s.size() && std::isspace(s[pos])) ++pos;
            if(pos >= s.size()) throw xml_error();
            
            if(s[pos] != '"' && s[pos] != '\'') throw xml_error();
            char terminator = s[pos++];
            
            std::string tagparamvalue;
            for(;;)
            {
                if(pos >= s.size()) throw xml_error();
                if(s[pos] == terminator) { ++pos; break; }
                tagparamvalue += ParseEntity(s, pos, utf8);
            }
            
            subtree->params.insert(std::make_pair(tagparamname, tagparamvalue));
            
            //std::wcerr << "Got param (" << tagparamname << ")(" << tagparamvalue << ")\n";
        }
        
        if(begin_slash)
        {
            throw xml_end_tag();
        }
        
        if(!found_trailing_slash && !meta_tag)
            ParseSubTree(*subtree, s, pos, KeepNS, utf8);
        
        if(!meta_tag)
            result.tags.insert(std::make_pair(tagname, subtree));
        else
        {
            for(XMLTree::parammap_t::const_iterator
                i = subtree->params.begin();
                i != subtree->params.end();
                ++i)
            {
                if(i->first == "encoding")
                {
                    std::string encoding = i->second;
                    if(UTF8_CompareCI(encoding, "utf-8"))
                    {
                        utf8 = true;
                    }
                    else
                    {
                        utf8 = false;
                    }
                }
            }
        }
    }
    
    void ParseSubTree(XMLTree& result, const std::string& s, unsigned& pos,
                      bool KeepNS, bool utf8)
    {
        std::string spaces;
        while(pos < s.size())
        {
            if(std::isspace(s[pos]))
            {
                if(!result.value.empty()) spaces += s[pos];
                ++pos;
                continue;
            }
            if(s[pos] == '<')
            {
                ++pos;
                try
                {
                    ParseTag(result, s, pos, KeepNS, utf8);
                }
                catch(xml_end_tag)
                {
                    return;
                }
                continue;
            }
            result.value += spaces;
            spaces = "";
            result.value += ParseEntity(s, pos, utf8);
        }
    }
}

const XMLTree ParseXML(const std::string& s, bool KeepNS)
{
    XMLTree result;
    try
    {
        unsigned pos=0;
        ParseSubTree(result, s, pos, KeepNS, true);
    }
    catch(xml_error err)
    {
        //FIXME: what now?
    }
    return result;
}

XMLTreeSet XMLTree::operator[] (const std::string& key) const
{
    XMLTreeSet result;
    
    typedef tagmap_t::const_iterator tagcit;
    std::pair<tagcit, tagcit> tmp = tags.equal_range(key);
    
    for(tagcit i=tmp.first; i!=tmp.second; ++i)
        result.AddTree(i->second);
    
    typedef parammap_t::const_iterator paramcit;
    std::pair<paramcit, paramcit> tmp2 = params.equal_range(key);
    
    for(paramcit i=tmp2.first; i!=tmp2.second; ++i)
        result.SetValue(i->second);

    return result;
}

const XMLTreeSet XMLTreeSet::operator[] (const std::string& key) const
{
    XMLTreeSet result;
    
    for(std::list<XMLTreeP>::const_iterator
        i = matching_trees.begin();
        i != matching_trees.end();
        ++i)
    {
        const XMLTree& p = *(*i);
        result.Combine( p[key] );
    }
    return result;    
}

void XMLTreeSet::AddTree(XMLTreeP p)
{
    matching_trees.push_back(p);
    matching_strings.push_back(p->value);
}

void XMLTreeSet::SetValue(const std::string& s)
{
    matching_strings.push_back(s);
}

void XMLTreeSet::Combine(const XMLTreeSet& b)
{
    matching_trees.insert(matching_trees.end(),     b.matching_trees.begin(),   b.matching_trees.end());
    matching_strings.insert(matching_strings.end(), b.matching_strings.begin(), b.matching_strings.end());
}

XMLTreeSet::operator const std::string () const
{
    if(matching_strings.empty()) return "";
    return *matching_strings.begin();
}

/*
XMLTreeSet::operator const XMLTreeP () const
{
    if(matching_trees.empty()) return new XMLTree;
    return *matching_trees.begin();
}
*/

void XMLTree::Dump(unsigned indent) const
{
    for(parammap_t::const_iterator i =params.begin(); i != params.end(); ++i)
    {
        printf("%*s",indent,"");
        printf("PARAM[%s]=%s\n", i->first.c_str(), i->second.c_str());
    }
    for(tagmap_t::const_iterator i = tags.begin(); i != tags.end(); ++i)
    {
        printf("%*s",indent,"");
        printf("TAG %s:\n", i->first.c_str());
        i->second->Dump(indent+2);
    }
    if(!value.empty())
    {
        printf("%*s",indent,"");
        printf("VALUE:\"%s\"\n", value.c_str());
    }
}

void XMLTreeSet::Dump() const
{
    for(std::list<std::string>::const_iterator
        i = matching_strings.begin(); i != matching_strings.end(); ++i)
    {
        printf("STRING:\"%s\"\n", i->c_str());
    }
    for(std::list<XMLTreeP>::const_iterator
        i = matching_trees.begin(); i != matching_trees.end(); ++i)
    {
        printf("TREE:\n");
        (*i)->Dump(2);
    }
}
